#include <stddef.h>
#include <stdio.h>
#include <errno.h>
#include <stdint.h>
#include <string.h>
#include <getopt.h>
#include <elf.h>
#include "../../kexec.h"
#include "../../kexec-syscall.h"
#include "kexec-sparc64.h"
#include <arch/options.h>

/* A huge WARNING here. Should the kernel linking change and there
 * be more than ONE PT_LOAD segment, then you have will HAVE work to do.
 */
#define MAX_MEMORY_RANGES	(128)
int max_memory_ranges;
static struct memory_range memory_range[MAX_MEMORY_RANGES];

int get_memory_ranges(struct memory_range **range, int *ranges,
		      unsigned long UNUSED(kexec_flags))
{
	int memory_ranges = 0;
	const char iomem[] = "/proc/iomem";
	char line[MAX_LINE];
	FILE *fp;
	unsigned long long start, end;
	char *str;
	int type, consumed, count;

	fp = fopen(iomem, "r");
	if (!fp) {
		fprintf(stderr, "Cannot open %s: %s\n", iomem, strerror(errno));
		return -1;
	}
	while (fgets(line, sizeof(line), fp) != 0) {
		if (memory_ranges >= MAX_MEMORY_RANGES)
			break;
		count = sscanf(line, "%Lx-%Lx : %n", &start, &end, &consumed);
		if (count != 2)
			continue;
		str = line + consumed;
		end = end + 1;
		if (memcmp(str, "System RAM\n", 11) == 0) {
			type = RANGE_RAM;
		} else if (memcmp(str, "reserved\n", 9) == 0) {
			type = RANGE_RESERVED;
		} else {
			continue;
		}
		memory_range[memory_ranges].start = start;
		memory_range[memory_ranges].end = end;
		memory_range[memory_ranges].type = type;
		memory_ranges++;
	}
	fclose(fp);
	*range = memory_range;
	*ranges = memory_ranges;
	max_memory_ranges = memory_ranges;
	return 0;
}

void kexec_move_pt_load(struct mem_ehdr *ehdr)
{
	int i;
	unsigned long start;
	unsigned long size = 0UL;

	for (i = 0; i < ehdr->e_phnum; i++) {
		struct mem_phdr *phdr = &ehdr->e_phdr[i];

		if (phdr->p_type != PT_LOAD)
			continue;
		else
			size += phdr->p_memsz;
	}

	for (i = 0; i < max_memory_ranges; i++)
		if ((memory_range[i].end - memory_range[i].start) >= size)
			break;

	if (i == max_memory_ranges)
		die("FUBAR\n");
	start = memory_range[i].start;

	for (i = 0; i < ehdr->e_phnum; i++) {
		struct mem_phdr *phdr = &ehdr->e_phdr[i];

		if (phdr->p_type != PT_LOAD)
			continue;
		else
			phdr->p_paddr = start + (phdr->p_vaddr - KERNBASE);
	}
}

static int elf_sparc64_probe(const char *buf, off_t len)
{
	struct mem_ehdr ehdr;
	int result;

	result = build_elf_exec_info(buf, len, &ehdr, 0);

	if (result < 0) {
		fprintf(stderr, "Not a ELF excutable\n");
		goto out;
	}

	if (ehdr.e_machine != EM_SPARCV9) {
		fprintf(stderr, "Not a sparc64 ELF\n");
		result = -1;
	}
out:
	return result;
}

static void elf_sparc64_usage(void)
{
	printf("initrd=FILE\n");
}

struct file_type file_type[] = {
	{"elf-sparc64", elf_sparc64_probe, elf_sparc64_load, elf_sparc64_usage},
};
int file_types = sizeof(file_type) / sizeof(file_type[0]);

void arch_usage(void)
{
}

int arch_process_options(int argc, char **argv)
{
	return 0;
}

const struct arch_map_entry arches[] = {
	{"sparc64", KEXEC_ARCH_SPARC64 },
	{ NULL, 0 },
};

int arch_compat_trampoline(struct kexec_info *UNUSED(info))
{

	return 0;
}

void arch_update_purgatory(struct kexec_info *UNUSED(info))
{
}

static char die_string[] = "Relocation not supported yet.\n";

void machine_apply_elf_rel(struct mem_ehdr *ehdr, unsigned long r_type,
	void *location, unsigned long address, unsigned long value)
{
	die(die_string);
}

int machine_verify_elf_rel(struct mem_ehdr *ehdr)
{
	die(die_string);
	return 0;
}
